@@ -2,13 +2,14 @@ package Tickit::Widget::Tree;
 # ABSTRACT: Simple tree widget for Tickit
 use strict;
 use warnings;
-use parent qw(Tickit::Widget);
+use parent qw(Tickit::ContainerWidget);
 use List::Util qw(sum);
 use Scalar::Util qw(blessed);
 use Tickit::Utils qw(substrwidth textwidth);
+use Tickit::Widget::Static;
 use utf8;
 
-our $VERSION = '0.002';
+our $VERSION = '0.003';
 
 =head1 NAME
 
@@ -16,19 +17,17 @@ Tickit::Widget::Tree - support for an expandable tree widget
 
 =head1 VERSION
 
-version 0.002
+version 0.003
 
 =head1 SYNOPSIS
 
  my $tree = Tickit::Widget::Tree->new(
  	is_open => 1,
- 	last => 1,
  	label => 'Root',
  	line_style => 'single',
  );
  $tree->add(Tickit::Widget::Tree->new(
  	is_open => 1,
- 	last => 1,
  	label => 'First',
  ));
 
@@ -72,24 +71,68 @@ sub new {
 	my $self = $class->SUPER::new(%args);
 	$self->{children} ||= [];
 
-	$args{last} ||= 1;
-
-	$self->{$_} = delete $args{$_} for grep exists $args{$_}, qw(label is_open last prev next line_style);
+	$self->{$_} = delete $args{$_} for grep exists $args{$_}, qw(label is_open prev next line_style label_widget);
 	$self->update_root_and_parent(
 		root	=> $root,
 		parent	=> $parent
 	);
+	$self->{pen_active} = delete $args{pen_active} || Tickit::Pen->new( fg => 2, bg => 4, b => 1 );
+	$self->{pen_label} = delete $args{pen_label} || Tickit::Pen->new( bg => 16, fg => 7 );
+
+	unless($self->{label_widget}) {
+		$self->{label_widget} = Tickit::Widget::Static->new(
+			text => defined($self->{label}) ? $self->{label} : ''
+		);
+		$self->{label_widget}->set_pen($self->pen_label);
+	}
+
 	if(my $children = delete $args{children}) {
-		foreach (@$children) {
-			$self->add(my $child = $class->new(%args, %$_) or die "fail?");
-			my $prefix_len = length($self->prefix_text);
-			# TODO - remember what was going to happen here
-		}
-		$self->resized;
+		$self->add(my $child = $class->new(%args, %$_) or die "failed to add child to $self") for @$children;
 	}
 	return $self;
 }
 
+=head2 label
+
+Set or retrieve the label for this widget.
+
+By default all nodes use a L<Tickit::Widget::Static> widget as a label,
+see the L</label_widget> method for more control over this.
+
+=cut
+
+sub label {
+	my $self = shift;
+	if(@_) {
+		my $txt = shift;
+		$self->{label} = $txt;
+		$self->label_widget->set_text($txt) if exists $self->{label_widget};
+		$self->resized;
+		return $self;
+	}
+	return $self->{label};
+}
+
+=head2 label_widget
+
+Set or retrieve the label widget for this node. Typically a L<Tickit::Widget::Static>,
+this must accept a pen which will be used for displaying the currently-highlighted node.
+
+=cut
+
+sub label_widget {
+	my $self = shift;
+	if(@_) {
+		$self->{label_widget} = shift;
+		$self->{label_widget}->set_pen($self->pen_label);
+		$self->resized;
+		return $self;
+	}
+	return $self->{label_widget};
+}
+
+# Is this still accurate?
+
 =head2 prev
 
 The 'previous element' is determined as follows:
@@ -132,7 +175,20 @@ sub prev {
 	return $self;
 }
 
+=head2 prev_sibling
+
+Returns the previous sibling for this node. RO accessor.
+
+=cut
+
 sub prev_sibling { shift->{prev_sibling} }
+
+=head2 next_sibling
+
+Returns the next sibling for this node. RO accessor.
+
+=cut
+
 sub next_sibling { shift->{next_sibling} }
 
 =head2 next
@@ -156,15 +212,22 @@ The 'next element' is determined as follows:
 sub next {
 	my $self = shift;
 	return $self->first_child if $self->is_open && $self->has_children;
-	return $self->next_ignore_children(@_);
+	return $self->_next_ignore_children(@_);
 }
 
-sub next_ignore_children {
+=head2 _next_ignore_children
+
+Returns the next node, ignoring any children in the current node for the
+purposes of the search.
+
+=cut
+
+sub _next_ignore_children {
 	my $self = shift;
 	my $ns = $self->next_sibling;
 	return $ns if $ns;
 	my $parent = $self->tree_parent or return $self;
-	my $suggested = $parent->next_ignore_children(@_);
+	my $suggested = $parent->_next_ignore_children(@_);
 	return $self if $suggested == $self->tree_parent;
 	return $suggested;
 }
@@ -178,7 +241,7 @@ unless the widget is open.
 
 sub lines {
 	my $self = shift;
-	my $lines = 1;
+	my $lines = $self->label_widget->lines;
 	return $lines unless $self->is_open;
 	$lines += sum 0, map +$_->lines, $self->children;
 	return $lines;
@@ -190,7 +253,7 @@ Number of cols for autocalculation. We're flexible, so we just use the default 1
 
 =cut
 
-sub cols { my $self = shift; 1; }
+sub cols { my $self = shift; textwidth($self->prefix_text) + $self->label_widget->cols; }
 
 =head2 is_open
 
@@ -208,27 +271,80 @@ sub is_open {
 	return 1;
 }
 
+=head2 is_closed
+
+Returns true if we're closed, false if not.
+
+=cut
+
 sub is_closed { !(shift->is_open) }
 
+=head2 window_lost
+
+Callback to clear children and label when we lose our window.
+
+=cut
+
 sub window_lost {
 	my $self = shift;
 
-	$_->set_window(undef) for $self->children;
-	$self->parent->resized if $self->parent;
+	$_->set_window(undef) for $self->label_widget, $self->children;
 	return $self;
 }
 
 =head2 window_gained
 
+Callback to adjust our widgets when we get a window.
+
 =cut
 
 sub window_gained {
 	my $self = shift;
 	# Let parent set up focus
 	$self->SUPER::window_gained(@_);
+	my $win = $self->window;
+	$win->clear;
+	$self->_reshape_label($win);
 	$self->reapply_windows;
 }
 
+=head2 reshape
+
+Callback to adjust the label window when our window is resized.
+
+=cut
+
+sub reshape {
+	my $self = shift;
+	$self->SUPER::reshape(@_);
+	my $win = $self->window;
+	$self->_reshape_label($win);
+}
+
+=head2 _reshape_label
+
+Adjust our label window to fit the current window.
+
+=cut
+
+sub _reshape_label {
+	my $self = shift;
+	my $win = shift;
+	die $self->label_widget->lines unless $self->label_widget->lines > 0;
+	my $prefix = $self->style_map->{solo};
+	my $prefix_len = textwidth $prefix;
+
+	# TODO try change_geometry again
+	$self->label_widget->set_window($win->make_sub(0, $prefix_len + 1, $self->label_widget->lines, $win->cols - ($prefix_len + 1)));
+	return $self;
+}
+
+=head2 reapply_windows
+
+Calculate windows for all widgets.
+
+=cut
+
 sub reapply_windows {
 	my $self = shift;
 
@@ -239,13 +355,12 @@ sub reapply_windows {
 
 	my $win = $self->window or return;
 
-	my $prefix_len = length $self->prefix_text;
-	my $y = 1;
+	my $prefix_len = textwidth $self->prefix_text;
+	my $y = $self->label_widget->lines;
 	my @tasks;
 	CHILD:
 	for my $child ($self->children) {
 		my $height = $child->lines;
-#		last CHILD unless ($y + $height) < $self->window->lines;
 
 		push @tasks, [ $child, ($child->window ? 'change' : 'create'), 
 			$y,
@@ -255,9 +370,10 @@ sub reapply_windows {
 		];
 		$y += $height;
 	}
-#	$win->resize($y, $win->cols);
+
 	foreach (@tasks) {
 		my ($child, $type, @args) = @$_;
+# TODO make this work again
 #		if($type eq 'create') {
 			my $sub = $win->make_sub(@args);
 			$child->set_window($sub);
@@ -298,6 +414,8 @@ sub insert_after {
 
 =head2 add
 
+Adds a child node after all other children.
+
 =cut
 
 sub add {
@@ -314,47 +432,91 @@ sub add {
 # Try to send the on_add signal
 	$_->($child, $self) for grep defined, map $child->can($_), qw(on_add);
 
-# Update last child to let it know that there's another widget following it now
-	$prev->{last} = 0 if $prev;
 	$self->resized;
 	$self;
 }
 
 =head2 children
 
+Returns a list of all children.
+
 =cut
 
 sub children { @{ shift->{children} || [] } }
 sub has_children { @{ shift->{children} || [] } ? 1 : 0 }
 sub first_child { shift->{children}[0] }
 sub last_child { shift->{children}[-1] }
+sub pen_label { shift->{pen_label} }
+sub pen_active { shift->{pen_active} }
 
 =head2 is_highlighted
 
+Returns true if this node is highlighted.
+
+Note that the highlighted widget itself is tracked through
+the root node, see L</highlighted> for more information.
+
 =cut
 
 sub is_highlighted { shift->{is_highlighted} }
 
 =head2 prefix_text
 
+Returns the appropriate line prefix text for the current style.
+
 =cut
 
 sub prefix_text {
 	my $self = shift;
-	my $style = $self->line_style;
-	if($style eq 'ascii') {
-		return ' | ';
-	} elsif($style eq 'compact_ascii') {
-		return '|';
-	} elsif($style eq 'single') {
-		return '│ ';
-	} elsif($style eq 'double') {
-		return ' ║';
-	} elsif($style eq 'thick') {
-		return ' ┃';
-	} else {
-		die "No line_style found for $self\n";
-	}
+	return $self->style_map->{prefix};
+}
+
+=head2 style_map
+
+Returns hashref which maps a component to the character string used to represent it.
+
+=cut
+
+sub style_map {
+	my $self = shift;
+	return {
+		ascii => {
+			prefix      => '| ',
+			is_open     => '[-]',
+			is_closed   => '[+]',
+			solo        => '-  ',
+			open_parent => '-+ ',
+			item_prefix => '+',
+			item_last   => '\\',
+		},
+		single => {
+			prefix      => '│ ',
+			is_open     => '[-]',
+			is_closed   => '[+]',
+			solo        => '─  ',
+			open_parent => '─┬ ',
+			item_prefix => '├',
+			item_last   => '└',
+		},
+		double => {
+			prefix      => '║ ',
+			is_open     => '[-]',
+			is_closed   => '[+]',
+			solo        => '═  ',
+			open_parent => '═╦ ',
+			item_prefix => '╠',
+			item_last   => '╚',
+		},
+		thick => {
+			prefix      => '┃ ',
+			is_open     => '[-]',
+			is_closed   => '[+]',
+			solo        => '━  ',
+			open_parent => '━┳ ',
+			item_prefix => '┣',
+			item_last   => '┗',
+		},
+	}->{$self->line_style};
 }
 
 =head2 render
@@ -367,59 +529,22 @@ sub render {
 	return unless my $win = $self->window;
 	my %args = @_;
 	my $rect = $args{rect} || $win->rect;
+	my $style_map = $self->style_map;
 
 	my $y = $rect->top;
 	if($y < 1) {
-		my $txt = '';
-		my $style = $self->line_style or die "No line style for $self";
-		if($style eq 'ascii') {
-			$txt = '[' . ($self->is_open ? '-' : '+') . ']';
-		} elsif($style eq 'compact_ascii') {
-			$txt = ($self->is_open ? '-' : '+');
-		} elsif($style eq 'single') {
-			$txt = $self->{last} ? '└' : '├'; # : '└');
-			if($self->children && $self->is_open) {
-				$txt .= '─┬ ';
-			} elsif($self->children) {
-				$txt .= '[' . ($self->is_open ? '-' : '+') . ']';
-			} else {
-				$txt .= '─  ';
-			}
-		} elsif($style eq 'double') {
-			$txt = $self->{last} ? '╚' : '╠';
-			if($self->children && $self->is_open) {
-				$txt .= '═╦ ';
-			} elsif($self->children) {
-				$txt .= '[' . ($self->is_open ? '-' : '+') . ']';
-			} else {
-				$txt .= '═  ';
-			}
-		} elsif($style eq 'thick') {
-			$txt = $self->{last} ? '┗' : '┣';
-			if($self->children && $self->is_open) {
-				$txt .= '━┳ ';
-			} elsif($self->children) {
-				$txt .= '[' . ($self->is_open ? '-' : '+') . ']';
-			} else {
-				$txt .= '   ';
-			}
+		my $txt = $style_map->{$self->next_sibling ? 'item_prefix' : 'item_last'};
+		if($self->children && $self->is_open) {
+			$txt .= $style_map->{open_parent};
+		} elsif($self->children) {
+			$txt .= $style_map->{'is_closed'};
 		} else {
-			die 'Invalid style for tree element';
+			$txt .= $style_map->{solo};
 		}
 		if($rect->left <= textwidth $txt) {
-			$win->goto(0,0);
+			$win->goto($y, $rect->left);
 			$win->print($txt);
 		}
-		my $lbl = $self->{label} // 'undef';
-		$lbl = substrwidth $lbl, 0, $win->cols;
-		if($rect->left <= textwidth "$txt$lbl") {
-			if($self->is_highlighted) {
-				$win->print($lbl, bg => 4);
-			} else {
-				$win->print($lbl);
-			}
-			$win->erasech($rect->cols - textwidth("$txt$lbl"));
-		}
 		++$y;
 	}
 
@@ -430,11 +555,12 @@ sub render {
 		my $x = 0;
 		$win->print($self->next_sibling ? $self->prefix_text : (' ' x textwidth($self->prefix_text))) if $rect->left <= textwidth($self->prefix_text);
 		$x += textwidth($self->prefix_text);
-		if($self->is_open && $self->children) {
-			$win->erasech($rect->cols - $x) unless $self->next_sibling;
-		} else {
-			$win->erasech($rect->cols - $x);
-		}
+		$win->erasech($rect->cols - $x, 1);
+#		if($self->is_open && $self->children) {
+#			$win->erasech($rect->cols - $x) unless $self->next_sibling;
+#		} else {
+#			$win->erasech($rect->cols - $x);
+#		}
 	}
 }
 
@@ -447,24 +573,52 @@ sub highlighted {
 	return $self->root->{highlighted} ||= $self->root;
 }
 
-=head2 label
-
-=cut
-
-sub label { shift->{label} }
-
 =head2 highlight
 
 =cut
 
 sub highlight {
 	my $self = shift;
-	$self->highlighted->{is_highlighted} = 0;
+	if(my $h = $self->highlighted) {
+		$h->_unhighlight;
+		$h->{is_highlighted} = 0;
+	}
+
 	$self->{is_highlighted} = 1;
+	$self->_highlight;
 	$self->root->{highlighted} = $self;
 	return $self;
 }
 
+sub _highlight {
+	my $self = shift;
+	return $self unless my $w = $self->label_widget;
+
+	$self->_set_pen_for_widget($w => 'active');
+	return $self;
+}
+
+sub _unhighlight {
+	my $self = shift;
+	return $self unless my $w = $self->label_widget;
+
+	$self->_set_pen_for_widget($w => 'label');
+	return $self;
+}
+
+sub _set_pen_for_widget {
+	my $self = shift;
+	my $w = shift or die "no widget supplied";
+	my $type = shift or die "no type supplied";
+
+	my $method = 'pen_' . $type;
+	$w->set_pen($self->$method);
+	# TODO This seems wrong, but somehow I couldn't convince Tickit::Widget to pick up the new pen
+	$w->window->set_pen($self->$method) if $w->window;
+	$w->resized;
+	return $self;
+}
+
 =head2 highlight_next
 
 =cut
@@ -567,31 +721,9 @@ sub on_key {
 	my ($type, $str, $key) = @_;
 	if($type eq 'key' && $str eq 'Down') {
 		$self->highlight_next;
-		$self->rerender;
 		return 1;
 	} elsif($type eq 'key' && $str eq 'Up') {
 		$self->highlight_prev;
-		$self->rerender;
-		return 1;
-	} elsif($type eq 'text' && $str eq 'A') {
-		my @pending = $self->root;
-		$self->window->focus(0,0) if $self->window;
-		while(@pending) {
-			my $node = shift @pending;
-			push @pending, $node->children;
-			$node->{is_open} = 1;
-		}
-		$self->root->resized;
-		return 1;
-	} elsif($type eq 'text' && $str eq 'a') {
-		my @pending = $self->root;
-		$self->window->focus(0,0) if $self->window;
-		while(@pending) {
-			my $node = shift @pending;
-			push @pending, $node->children;
-			$node->{is_open} = 0;
-		}
-		$self->root->resized;
 		return 1;
 	} elsif($type eq 'text' && $str eq '+') {
 		$self->highlighted->open;
@@ -603,22 +735,14 @@ sub on_key {
 		$self->root->rerender;
 		return 1;
 	}
-	if($type eq 'key' && $str eq 'Tab') {
-		my $target;
-		unless($target) {
-			$target = $self;
-			$target = $target->parent while $target->parent;
-		}
-		$target->window->focus(0,0) if $target->window;
-		$target->children_changed if $target->can('children_changed');
-		return 1;
-	}
 	return $self->_updated_on_key(@_);
 	return 0;
 }
 
 =head2 open
 
+Open this node. Triggers a redraw.
+
 =cut
 
 sub open {
@@ -626,11 +750,14 @@ sub open {
 	$self->{is_open} = 1;
 	$self->tree_parent->reapply_windows if $self->tree_parent;
 	$self->next_sibling->reapply_windows if $self->next_sibling;
+	$self->resized;
 	return $self;
 }
 
 =head2 close
 
+Close this node. Triggers a redraw.
+
 =cut
 
 sub close {
@@ -642,16 +769,8 @@ sub close {
 	return $self;
 }
 
-=head2 child_resized
-
-=cut
-
 sub child_resized { }
 
-=head2 bind_keys
-
-=cut
-
 sub bind_keys {
 	my $self = shift;
 	while(@_) {
@@ -696,18 +815,22 @@ sub _updated_on_key {
 
 =head2 on_add
 
+Make sure everything has a consistent view of root and parent nodes
+when adding a widget to a parent.
+
 =cut
 
 sub on_add {
 	my $self = shift;
 	my $parent = shift;
 	$self->update_root_and_parent(parent => $parent);
-	$self->{last} = 1;
 	return $self;
 }
 
 =head2 line_style
 
+Returns current line style. Inherits from parent, root or defaults to 'single'.
+
 =cut
 
 sub line_style {
@@ -724,21 +847,6 @@ sub line_style {
 	return 'single';
 }
 
-=head2 is_valid_tree_node
-
-Returns 1 if this is a valid tree node, 0 if not.
-
-A 'valid tree node' is defined as an object which is a subclass or instance
-of the L<Tickit::Widget::Tree> class.
-
-=cut
-
-sub is_valid_tree_node {
-	my $self = shift;
-	my $node = shift;
-	return +(blessed($node) && $node->isa(__PACKAGE__)) ? 1 : 0;
-}
-
 1;
 
 __END__
